*********************************************************************
* ZZError.prg
PARAMETERS nErrNum,cErrMsg,cErrPrg,cErrLine,nErrLine
*
* SalesBridge error handler
* Adapted from Inside FoxPro 2.5 for Windows -- Chapter 9
* See "Programmer's Notes" in SalesBridge help for more information
*********************************************************************

PRIVATE cConsole,cDevice,cPrint,cWinSave,cChoice,nKey, ;
        cLine,cEhSave,bOWVis,cTalk,n,nArea,cMemory,cStatus

cEhSave = ON('ERROR')

ON ERROR WAIT 'Invalid command, try again.' WINDOW TIMEOUT 1.5

IF nErrNum = 1707
  * Structural CDX not found.  RETRY will unmark the DBF.
  RETRY
ENDIF

IF SET('TALK') = 'ON'
  SET TALK OFF
  cTalk = 'ON'
ELSE
  cTalk = 'OFF'
ENDIF

cConsole = SET('CONSOLE')
cDevice  = SET('DEVICE')
cPrint   = SET('PRINTER')
SET CONSOLE ON
SET PRINT OFF
SET DEVICE TO SCREEN
cWinSave = WOUTPUT()
bOWVis = WVISIBLE(cWinSave)
PUSH KEY CLEAR

IF nErrNum = 125
  * Printer not ready.  Give them "Printer not ready.  Retry?" message.
  IF PRCHECK()
    DO RESTENV
    
    RETRY
  ELSE
    * The following would work fine with the report writer, since
    * it would resume executing right after the REPORT command.
    * For hand-coded reports you'd have to RETURN TO the program
    * calling the report or perform an abort.
    DO RESTENV
    RETURN
  ENDIF
ENDIF

IF cErrLine = cErrMsg
  * If using the runtime, MESSAGE(1) sometimes returns the same as
  * MESSAGE(), so supress displaying the same thing twice.
  cErrLine = ''
ENDIF

cErrLine = ALLTRIM(cErrLine)
cErrMsg = ALLTRIM(cErrMsg)
IF NOT EMPTY(cErrPrg)
  IF LEFT(cErrPrg,1) = '_' OR '...' $ cErrPrg
    * Error generated by a machine-named routine; show the first parent
    * routine with a hand-coded name also
    cPS = ' (' + MYPROG(2) + ')'
  ELSE
    cPS = ''
  ENDIF
  cErrPrg = 'PROBLEM in program ' + ALLTRIM(cERRPRG) + cPS
ENDIF
cLine = IIF(nErrLine = 0,'','  (Line ' + LTRIM(STR(nErrLine,5,0)) + ')')
cErrLine = ALLTRIM(cErrLine)
cErrLine1 = RTRIM(LEFT(cErrLine,63))

IF LEN(cErrLine) > 63
  cErrLine2 = RTRIM(SUBSTR(cErrLine,64,63))
ELSE
  cErrLine2 = ''
ENDIF

IF DISKSPACE() >= 204800

  * Log the error if 200k or more of disk space available

  * Save the diagnostic dumps first, before we add new
  * memory variables and (potentially) close a table
  * and open zzerror table in subsequent code

  cMemory = SYS(3) + '.tmp'
  SET CONSOLE OFF
  SET DEVICE TO FILE (cMemory)
  LIST MEMORY NOCONSOLE TO FILE (cMemory)
  cStatus = SYS(3) + '.tmp'
  SET DEVICE TO FILE (cStatus)
  LIST STATUS NOCONSOLE TO FILE (cStatus)
  SET DEVICE TO SCREEN
  SET CONSOLE ON

*susp

  IF nErrNum = 1721
    * The error is "All work areas are in use". This is fatal
    * anyway, so close one table so we can have a work
    * area available to log the error.
    USE IN 1
  ENDIF

  nArea = SELECT()
  IF NOT FILE('zzerror.dbf')
    SELECT 0
    CREATE TABLE zzerror ;
                (errnum    N(4,0), ;
                 date      D,      ;
                 time      C(8),   ;
                 errprg    M,      ;
                 code      M,      ;
                 errmsg    M,      ;
                 status    M,      ;
                 memory    M,      ;
                 traceback M)
    USE
  ENDIF
  SELECT 0
  USE zzerror SHARED
  IF LOCKF()
    APPEND BLANK
    UNLOCK
    IF LOCKR()
      REPLACE errnum WITH nErrNum, ;
              errmsg WITH cErrMsg, ;
              date   WITH DATE(),  ;
              time   WITH TIME(),  ;
              errprg WITH cErrPrg, ;
              code   WITH cErrLine
      APPEND MEMO status FROM (cStatus)
      APPEND MEMO memory FROM (cMemory)
      n = 1
      DO WHILE n <= 32 AND NOT EMPTY(PROGRAM(n))
        REPLACE traceback WITH ;
                traceback + STR(n,2,0) + ' ' + SYS(16,n) ;
                          + CHR(13) + CHR(10)
        n = n + 1
      ENDDO
    ENDIF
  ENDIF
  USE
  SELECT (nArea)
  ERASE (cStatus)
  ERASE (cMemory)
ENDIF

DO WHILE .t.

  IF nErrNum = 1249 OR RDLEVEL() = 5
    * Too many READs in effect -- zzalert() can't use buttons, since
    * that would invoke another READ level
    DO zzalert WITH 'STOP','',-1,-1,;
                  'PROBLEM' + cErrPrg + cLine, ;
                  cErrMsg, cErrLine1,cErrLine2
    IF nErrNum = 1249
      WAIT 'Press any key to abort' WINDOW
      cChoice = 'A'
      RELEASE WINDOW zzalert
    ELSE
      DEFINE WINDOW notify FROM 1,SCOLS() - 50 TO 3,SCOLS() - 1 ;
                    COLOR SCHEME 5
      ACTIVATE WINDOW notify
      @ 0,1 SAY 'Press [I]gnore, [R]etry, [S]uspend, or [A]bort'
      nKey = 0
      CLEAR TYPEAHEAD
      DO WHILE nKey <> ASC('I') AND ;
               nKey <> ASC('i') AND ;
               nKey <> ASC('R') AND ;
               nKey <> ASC('r') AND ;
               nKey <> ASC('S') AND ;
               nKey <> ASC('s') AND ;
               nKey <> ASC('A') AND ;
               nKey <> ASC('a') AND ;
               nKey <> 27 AND ;
               nKey <> 10
        nKey = INKEY(0)
      ENDDO
      RELEASE WINDOW notify
      DO CASE
        CASE nKey = 27
          * Esc = Retry
          cChoice = 'R'
        CASE nKey = 10
          * Ctrl+Enter = Abort
          cChoice = 'A'
        OTHERWISE
          cChoice = UPPER(CHR(nKey))
      ENDCASE
    ENDIF
  ELSE
    IF nErrNum = 1721
      * All work areas used -- always fatal, especially since
      * error logging routine probably closed all tables anyway
      DO zzalert WITH 'STOP','\!\?\<Abort',-1,-1,;
                    cErrPrg + cLine, cErrMsg,cErrLine1,cErrLine2
      cChoice = 'A'
    ELSE
      IF _Ok2All
        cChoice = zzalert('STOP','\<Ignore;\<Suspend;\!\<ReStart',-1,-1, ;
                  cErrPrg + cLine, cErrMsg, cErrLine1, cErrLine2)
      ELSE
        cChoice = zzalert('STOP','\<Ignore;\!\<ReStart',-1,-1, ;
                  cErrPrg + cLine, cErrMsg, cErrLine1, cErrLine2)
      ENDIF
    ENDIF
  ENDIF

  DO CASE
    CASE cChoice = 'I'
      * Ignore
      DO RESTENV
      RETURN

    CASE cChoice = 'C'
      * Cancel
      DO wShutDown
      RETURN TO MASTER
      
    CASE cChoice = 'S'

      IF NOT _Ok2All
        WAIT 'FoxPro command window, Access Denied!' WINDOW TIMEOUT 1.5
      ENDIF
      IF _Ok2All
        * assume running under the runtime, which doesn't support suspend, so
        * we simulate it

        * Not working, can't seem to reset "on error" setting
        * from within the "on error" routine specified! pds 11/94
        ON ERROR WAIT 'Invalid command, try again.' WINDOW TIMEOUT 1.5

        SET CONSOLE ON
        SET PRINT OFF
        SET DEVICE TO SCREEN

        DEFINE WINDOW cmd FROM 0,0 TO SROWS() - 4, 72 FLOAT GROW SYSTEM ;
                      TITLE 'Enter FoxPro commands   "RESU" "RESTART" '
        ACTIVATE WINDOW cmd

        DO WHILE .t.

          ON ERROR WAIT 'Invalid command, try again.' WINDOW TIMEOUT 1.5

          cCMD = ''
          ACCEPT '. ' TO cCMD
          cCMD = UPPER(cCMD)

          DO CASE
          CASE EMPTY(cCMD)
            WAIT 'No command entered, try again.' WINDOW TIMEOUT 1.5
            LOOP
          CASE ALLTRIM(cCMD) = 'RESTART'
            * ReStart (warm boot entire app)
            DO RESTENV            && restore environment
            SET CONSOLE ON        && set up initial conditions for "warm boot"
            SET PRINT OFF
            SET DEVICE TO SCREEN
            CLOSE DATABASES
            SET EXCLUSIVE OFF     && assuming this is a multiuser app
            CLEAR WINDOWS
            CLEAR POPUPS
            CLEAR READ ALL        && cancel all READs, including foundation READ

            * (to re-paint main Fox screen and exit main loop in zz.prg)
            _SvWall = 'exit' + _Wall
            * Refer to top level program by name
            IF RIGHT(CALLER(),3) = 'ZZ'
              cACTION = 'RETURN TO MASTER'
              RETURN
            ELSE
              RETURN TO MASTER
            ENDIF

          CASE ALLTRIM(cCMD) = 'SUSP'
            ON ERROR
            SUSP
            ON ERROR WAIT 'Invalid command, try again.' WINDOW TIMEOUT 1.5
            LOOP
          CASE ALLTRIM(cCMD) = 'CANC'
            WAIT '"Cancel" not allowed here, "RESTART" then cancel from SalesBridge menu.' WINDOW TIMEOUT 1.5
            LOOP
          CASE ALLTRIM(cCMD) = 'QUIT'
            WAIT '"Quit" not allowed here, quit from SalesBridge menu.' WINDOW TIMEOUT 1.5
            LOOP
          CASE EMPTY(cCMD) OR ALLTRIM(cCMD) = 'RESU' ;
               OR ALLTRIM(cCMD) = 'RETURN'
            * Ignore
            DO RESTENV
            EXIT
          OTHERWISE
            * execute FoxPro commands, no way to turn off Fox Errors here,
            * can't seem to reset "on error" setting
            * from within the "on error" routine specified! pds 11/94
            * 
            &cCMD
          ENDCASE
        ENDDO {.t.}

        ON ERROR DO zzerror WITH ERROR(),MESSAGE(),PROGRAM(),MESSAGE(1),LINENO(1)

      ENDIF { Ok2All }

    CASE cChoice = 'R'
      * ReStart (warm boot entire app)
      DO RESTENV            && restore environment
      SET CONSOLE ON        && set up initial conditions for "warm boot"
      SET PRINT OFF
      SET DEVICE TO SCREEN
      CLOSE DATABASES
      SET EXCLUSIVE OFF     && assuming this is a multiuser app
      CLEAR WINDOWS
      CLEAR POPUPS
      CLEAR READ ALL        && cancel all READs, including foundation READ

      * (to re-paint main Fox screen and exit main loop in zz.prg)
      _SvWall = 'exit' + _Wall
      * Refer to top level program by name
      IF RIGHT(CALLER(),3) = 'ZZ'
        cACTION = 'RETURN TO MASTER'
        RETURN
      ELSE
        RETURN TO MASTER
      ENDIF

  ENDCASE

ENDDO

* EOF - zzerror
***************************************************************
FUNCTION PRCHECK
* Checks printer status, informs operator if offline.  Operator may correct
* problem or abort.  Returns false if operator abort, true otherwise.
*

DO WHILE NOT PRINTSTATUS()
  IF zzalert('STOP','Yn',-1,-1,'Printer not ready.  Retry?') = 'N'
    RETURN .f.
  ENDIF
ENDDO

RETURN .t.

* EOF - PrCheck()
***************************************************************
FUNCTION CALLER
* This function returns the name of the program that called the routine that
* invokes this function.  If optional numeric argument is passed, the
* that many additional DO nesting levels are subtracted.  Example:
* If the current DO nesting is:
*   MAIN
*   MENU1
*   SCREEN1
*   _P0V1OX2B  <-- CALLER() executed here
*
* Then CALLER() would return "SCREEN1"
*      CALLER(1) would return "MENU1"
*      CALLER(2) would return "MAIN"
*      CALLER(3) would return ""

PARAMETER nLevel

PRIVATE n
* Find out how deep we are in the stack
n = 1

DO WHILE PROGRAM(n) <> 'CALLER'
  n = n + 1
ENDDO

* The following EMPTY(nLevel) test deals with the possibility that nLevel
* may actually be logical false because no arguments were passed;
* it's quicker than using TYPE() and effectively defaults nLevel to zero.

IF EMPTY(nLevel)
  RETURN IIF(n < 2,'',PROGRAM(n - 2))
ELSE
  n = n - 2 - nLevel
  RETURN IIF(n < 0,'',PROGRAM(n))
ENDIF

* EOF - Caller
***************************************************************
FUNCTION MYPROG
* This function is identical to PROGRAM(), except that if the currently
* executing program has a program name generated by the system, we work
* our way back up the calling chain until we find a program with a human-
* generated name.  This is because names created by program generators
* change each time the screen or menu or whatever are regenerated.
* Also, we refuse to return "ON..."

* Note that at runtime we are one level deeper in the DO stack (deeper
* in DO??) than where the actual MYPROG() call is.  Thus at runtime we look
* at our caller's name to see the program executing at the time MYPROG()
* was called, and work backward from there.  Otherwise this function would
* always return "MYPROG".

* The search can optionally begin nHowFar levels above the default current
* nesting level.

PARAMETER nHowFar

IF PARAMETERS() = 0
  nHowFar = 0
ENDIF
PRIVATE cCaller,n
cCaller = CALLER(nHowFar)

IF LEFT(cCaller,1) = '_' OR '...' $ cCaller
  n = 1 + nHowFar
  DO WHILE NOT EMPTY(cCaller)
    cCaller = CALLER(n)
    IF LEFT(cCaller,1) <> '_' AND NOT '...' $ cCaller
      EXIT
    ENDIF
    n = n + 1
  ENDDO
ENDIF

RETURN cCaller


* EOF - MyProg
***************************************************************
FUNCTION LOCKF
*
*    This function attempts to lock the current table.  The user is prompted
*    if the table cannot be locked.
*

IF FLOCK()
  RETURN .t.
ENDIF

PRIVATE nInKey

DO zzalert WITH 'INFO','',-1,-1, ;
              [ File "] + DBF() + [" in use...], ;
              "Waiting for access, Press [Esc] to cancel"
CLEAR TYPEAHEAD
nInKey = INKEY(1)
DO WHILE nInKey <> 27 AND (NOT FLOCK())
  nInKey = INKEY(1)
ENDDO
RELEASE WINDOW zzalert

RETURN nInKey <> 27

* EOF - LockF
***************************************************************
FUNCTION LOCKR
*
*    This function will attempt to lock the current record in the current
*    table.  The user is prompted if the record cannot be locked.
*

IF RLOCK()
  RETURN .t.
ENDIF

PRIVATE nInKey

DO zzalert WITH 'INFO','',-1,-1, ;
              [Record in file "] + DBF() + [" in use...], ;
              '   Waiting for access, Press [Esc] to cancel'

CLEAR TYPEAHEAD
nInKey = INKEY(1)
DO WHILE nInKey <> 27 AND (NOT RLOCK())
  nInKey = INKEY(1)
ENDDO

RELEASE WINDOW zzalert

RETURN nInKey <> 27

* EOF - LockR
***************************************************************
PROCEDURE RESTENV
*
* called by error handler to restore environment in existence before error
* handler was invoked

SET CONSOLE &cConsole
SET DEVICE TO &cDevice
SET PRINT &cPrint
RELEASE WINDOW cmd
IF EMPTY(cWinSave)
  ACTIVATE SCREEN
ELSE
  IF bOWVis
    ACTIVATE WINDOW (cWinSave) SAME
  ELSE
    ACTIVATE WINDOW (cWinSave) SAME NOSHOW
  ENDIF
ENDIF
POP KEY
IF ON('ERROR') <> cEhSave
  ON ERROR &cEhSave
ENDIF
IF cTalk = 'ON'
  SET TALK ON
ENDIF

RETURN

* EOF - RestEnv
***************************************************************
***************************************************************
***************************************************************
*EOF - zzError.prg



* old code, save for future reference, pds 11/94
do while .f.
@  1, 1 SAY 'Message'
@  1,22 SAY SUBSTR(errormsg,1,45)
@  2, 1 SAY 'Program Name'
@  2,22 SAY prgname
@  3, 1 SAY 'Program Line #'
@  3,22 SAY ALLTRIM(STR(LINENO))
@  4, 1 SAY 'Current Field'
@  4,22 SAY SYS(18)
@  5, 1 SAY 'Disk Space'
@  5,22 SAY TRANSFORM(DISKSPACE(),'#,###,###,###')
@  6, 1 SAY 'Database In Use'
@  6,22 SAY ALIAS()
@  7, 1 SAY 'Directory'
@  7,22 SAY SYS(5)+sys(2003)

@  9, 1 SAY 'Message Number'
@  9,22 SAY ALLTRIM(STR(errnum))
@ 10, 1 SAY 'Operating Syst'
@ 10,22 SAY OS()
@ 11, 1 SAY 'CPU'
@ 11,22 SAY SYS(17)
@ 12, 1 SAY 'CONFIG.SYS Files'
@ 12,22 SAY SYS(2010)
@ 13, 1 SAY 'Graphics Card'
@ 13,22 SAY SYS(2006)
@ 14, 1 SAY 'Network Machine'
@ 14,22 SAY ALLTRIM(SYS(0))

@ 11,38 SAY 'Mem Avail/640K'
@ 11,59 SAY TRANSFORM(VAL(SYS(12)),'#,###,###')
@ 12,38 SAY 'Total Mem Avail'
@ 12,59 SAY TRANSFORM(VAL(SYS(1001)),'#,###,###')
@ 13,38 SAY 'EMS Left Over'
@ 13,59 SAY SYS(23) + ' K'
@ 14,38 SAY 'Config EMS Limit'
@ 14,59 SAY SYS(24) + ' K'

enddo